from builtins import str
from builtins import object
import logging
import copy
from elasticsearch import Elasticsearch


class BmajIndex(object):
    """
    ElasticSearch indexation and search
    """

    """
    ElasticSearch server
    """
    es = None

    """
    Index name
    """
    index = 'biomaj'

    """
    Do indexing
    """
    do_index = False

    """
    Skip if failure (tests)
    """
    skip_if_failure = False

    @staticmethod
    def load(hosts=None, index='biomaj', do_index=True):
        """
        Initialize index

        :param hosts: List of elastic search nodes to connect to
        :type hosts: list
        :param do_index: index data or not
        :type do_index: bool
        """
        if hosts is None:
            hosts = ['localhost']
        if not do_index:
            return
        BmajIndex.index = index
        BmajIndex.do_index = do_index
        if BmajIndex.es is None:
            BmajIndex.es = Elasticsearch(hosts)

            mapping = {
                "mappings": {
                    "production": {
                        "date_detection": False
                    },
                    "releasestats": {
                        "date_detection": False,
                    }
                }
            }
            try:
                if not BmajIndex.es.indices.exists(index=BmajIndex.index):
                    BmajIndex.es.indices.create(index=BmajIndex.index, body=mapping)
            except Exception as e:
                logging.error('ElasticSearch connection error, check server is running and configuration: ' + str(e))

    @staticmethod
    def _bulk_delete(query, flush=True):
        try:
            page = BmajIndex.es.search(
                index=BmajIndex.index,
                doc_type='production',
                search_type="query_then_fetch",
                size=1000,
                body={
                    'query': {
                        'match': {
                            'bank': query['bank']
                        }
                    }
                })

            if page is None:
                return
            bulk_delete = ''
            for del_hit in page['hits']['hits']:
                if ('release' in query and query['release'] == del_hit['_source']['release']) or 'release' not in query:
                    bulk_delete += "{ \"delete\" : {\"_index\":\"" + BmajIndex.index + "\",\"_type\":\"production\", \"_id\" : \"" + del_hit['_id'] + "\" } }\n"
            if bulk_delete:
                BmajIndex.es.bulk(body=bulk_delete)
                if flush:
                    BmajIndex.es.indices.flush(index=BmajIndex.index, force=True)
        except Exception as e:
            if BmajIndex.skip_if_failure:
                BmajIndex.do_index = False
            else:
                logging.error('Failed to delete bank index: ' + str(e))

    @staticmethod
    def delete_all_bank(bank_name):
        """
        Delete complete index for a bank
        """
        if not BmajIndex.do_index:
            return
        BmajIndex._bulk_delete({"bank": bank_name}, True)

    @staticmethod
    def remove(bank_name, release):
        """
        Remove a production release

        :param bank_name: Name of the bank
        :type bank_name: str
        :param release: production release
        :type release: str
        """
        if not BmajIndex.do_index:
            return
        BmajIndex._bulk_delete({"release": release, "bank": bank_name})
        """
        try:
            query = {
              "query" : {
                "term" : {"release" : release, "bank": bank_name}
                }
              }
            BmajIndex.es.delete_by_query(index=BmajIndex.index, body=query)
        except Exception as e:
            logging.error('Index:Remove:'+bank_name+'_'+str(release)+':Exception:'+str(e))
            if BmajIndex.skip_if_failure:
                BmajIndex.do_index = False
        """

    @staticmethod
    def search(query):
        if not BmajIndex.do_index:
            return None
        res = BmajIndex.es.search(index=BmajIndex.index,
                                  doc_type='production',
                                  search_type="query_then_fetch",
                                  body=query)
        return res['hits']['hits']

    @staticmethod
    def searchq(query, size=1000):
        """
        Lucene syntax search

        :param query: Lucene search string
        :type query: str
        :param size: number of results
        :type size: int
        :return: list of matches
        """
        if not BmajIndex.do_index:
            return None
        res = BmajIndex.es.search(index=BmajIndex.index, doc_type='production', q=query, size=size)
        return res['hits']['hits']

    @staticmethod
    def add_stat(stat_id, stat):
        """
        Add some statistics, must contain release and bank properties.
        """
        if not BmajIndex.do_index:
            return
        if stat['release'] is None or stat['bank'] is None:
            return False
        try:
            BmajIndex.es.index(index=BmajIndex.index, doc_type='releasestats', id=stat_id, body=stat)
        except Exception:
            if BmajIndex.skip_if_failure:
                BmajIndex.do_index = False
            else:
                return False
        return True

    @staticmethod
    def add(bank_name, prod, flush=False):
        """
        Index a production release

        :param bank_name: Name of the bank
        :type bank_name: str
        :param prod: session release object
        :type prod: dict
        :param flush: Force flushing
        :type flush: bool
        """
        if not BmajIndex.do_index:
            return
        obj = copy.deepcopy(prod)
        if obj['release'] is None:
            return
        obj['bank'] = bank_name
        formats = obj['formats']
        try:
            for fkey, fvalue in formats.items():
                for elt in fvalue:
                    elt['format'] = fkey
                    elt['bank'] = bank_name
                    elt['release'] = obj['release']
                    if 'status' in obj:
                        elt['status'] = obj['status']
                    BmajIndex.es.index(index=BmajIndex.index, doc_type='production', body=elt)
            if flush:
                BmajIndex.es.indices.flush(index=BmajIndex.index, force=True)
        except Exception as e:
            logging.error('Index:Add:' + bank_name + '_' + str(obj['release']) + ':Exception:' + str(e))
            if BmajIndex.skip_if_failure:
                BmajIndex.do_index = False



from __future__ import print_function
from future import standard_library
from builtins import str
from builtins import object
import logging
import logging.config
import os
import time
import sys

from biomaj_core.bmajindex import BmajIndex

if sys.version < '3':
    import ConfigParser as configparser
else:
    import configparser

standard_library.install_aliases()


class BiomajConfig(object):
    """
    Manage Biomaj configuration
    """

    DEFAULTS = {
        'http.parse.dir.line': r'<img[\s]+src="[\S]+"[\s]+alt="\[DIR\]"[\s]*/?>[\s]*<a[\s]+href="([\S]+)/"[\s]*>.*([\d]{2}-[\w\d]{2,5}-[\d]{4}\s[\d]{2}:[\d]{2})',
        'http.parse.file.line': r'<img[\s]+src="[\S]+"[\s]+alt="\[[\s]+\]"[\s]*/?>[\s]<a[\s]+href="([\S]+)".*([\d]{2}-[\w\d]{2,5}-[\d]{4}\s[\d]{2}:[\d]{2})[\s]+([\d\.]+[MKG]{0,1})',
        'http.group.dir.name': 1,
        'http.group.dir.date': 2,
        'http.group.file.name': 1,
        'http.group.file.date': 2,
        'http.group.file.size': 3,
        'visibility.default': 'public',
        'historic.logfile.level': 'INFO',
        'bank.num.threads': 2,
        'files.num.threads': 4,
        'use_elastic': 0,
        'use_drmaa': 0,
        'db.type': '',
        'db.formats': '',
        'keep.old.version': 1,
        'docker.sudo': '1',
        'auto_publish': 0
    }

    # Old biomaj level compatibility
    LOGLEVEL = {
        'DEBUG': logging.DEBUG,
        'VERBOSE': logging.INFO,
        'INFO': logging.INFO,
        'WARN': logging.WARNING,
        'ERR': logging.ERROR
    }

    """
    Global configuration file
    """
    global_config = None

    """
    Per use global configuration file, overriding global_config
    """
    user_config = None
    allow_user_config = False

    @staticmethod
    def load_config(config_file=None, allow_user_config=True):
        """
        Loads general config

        :param config_file: global.properties file path
        :type config_file: str
        :param allow_user_config: use ~/.biomaj.cfg if present
        :type allow_user_config: bool
        """
        BiomajConfig.allow_user_config = allow_user_config

        if config_file is None:
            env_file = os.environ.get('BIOMAJ_CONF')
            if env_file is not None and os.path.exists(env_file):
                config_file = env_file
            else:
                env_file = 'global.properties'
                if os.path.exists(env_file):
                    config_file = env_file

        if config_file is None or not os.path.exists(config_file):
            raise Exception('Missing global configuration file')

        BiomajConfig.config_file = os.path.abspath(config_file)

        BiomajConfig.global_config = configparser.ConfigParser()

        # if allow_user_config and os.path.exists(os.path.expanduser('~/.biomaj.cfg')):
        #    BiomajConfig.user_config_file = os.path.expanduser('~/.biomaj.cfg')
        #    BiomajConfig.user_config = configparser.ConfigParser()
        #    BiomajConfig.user_config.read([os.path.expanduser('~/.biomaj.cfg')])
        # else:
        #    BiomajConfig.user_config_file = None

        BiomajConfig.global_config.read([config_file])

        # ElasticSearch indexation support
        do_index = False
        if BiomajConfig.global_config.get('GENERAL', 'use_elastic') and \
                BiomajConfig.global_config.get('GENERAL', 'use_elastic') == "1":
            do_index = True
        if do_index:
            if BiomajConfig.global_config.get('GENERAL', 'elastic_nodes'):
                elastic_hosts = BiomajConfig.global_config.get('GENERAL',
                                                               'elastic_nodes')\
                    .split(',')
            else:
                elastic_hosts = ['localhost']
            elastic_index = BiomajConfig.global_config.get('GENERAL', 'elastic_index')
            if elastic_index is None:
                elastic_index = 'biomaj'

            if BiomajConfig.global_config.has_option('GENERAL', 'test') and \
                    BiomajConfig.global_config.get('GENERAL', 'test') == "1":
                # Test connection to elasticsearch, if not available skip indexing for tests
                BmajIndex.skip_if_failure = True

            BmajIndex.load(index=elastic_index,
                           hosts=elastic_hosts,
                           do_index=do_index)

    def __init__(self, bank, options=None):
        """
        Loads bank configuration

        :param bank: bank name
        :type bank: str
        :param options: bank options
        :type options: argparse
        """
        self.log_level = logging.INFO
        self.name = bank
        if BiomajConfig.global_config is None:
            BiomajConfig.load_config()
        self.config_bank = configparser.ConfigParser()
        conf_dir = BiomajConfig.global_config.get('GENERAL', 'conf.dir')
        if not os.path.exists(os.path.join(conf_dir, bank + '.properties')):
            logging.error('Bank configuration file does not exists')
            raise Exception('Configuration file ' + bank +
                            '.properties does not exists')
        try:
            config_files = [BiomajConfig.config_file]
            self.user_config_file = None
            if self.allow_user_config and hasattr(options, 'user')\
                    and options.user\
                    and os.path.exists(os.path.expanduser('~' + options.user + '/.biomaj.cfg')):
                self.user_config_file = os.path.expanduser('~/' + options.user +
                                                           '.biomaj.cfg')
                self.user_config = configparser.ConfigParser()
                self.user_config.read([os.path.expanduser('~/' + options.user +
                                                          '.biomaj.cfg')])

            if self.user_config_file is not None:
                config_files.append(self.user_config_file)
            config_files.append(os.path.join(conf_dir, bank + '.properties'))
            self.config_bank.read(config_files)
        except Exception as e:
            logging.error("Configuration file error " + str(e))
            raise Exception("Configuration file error " + str(e))

        self.last_modified = int(os.stat(os.path.join(conf_dir, bank +
                                                      '.properties')).st_mtime)

        do_log = False
        if options is None:
            do_log = True
        elif hasattr(options, 'no_log') and not options.no_log:
            do_log = True
        elif type(options) is dict and 'no_log' in options and not options['no_log']:
            do_log = True

        if do_log:
            if os.path.exists(os.path.expanduser('~/.biomaj.cfg')):
                logging.config.fileConfig(os.path.expanduser('~/.biomaj.cfg'))
            else:
                logging.config.fileConfig(BiomajConfig.config_file)

            logger = logging.getLogger()
            bank_log_dir = os.path.join(self.get('log.dir'), bank, str(time.time()))
            if not os.path.exists(bank_log_dir):
                os.makedirs(bank_log_dir)
            hdlr = logging.FileHandler(os.path.join(bank_log_dir, bank + '.log'))
            self.log_file = os.path.join(bank_log_dir, bank + '.log')
            if options is not None and options.get_option('log') is not None:
                hdlr.setLevel(BiomajConfig.LOGLEVEL[options.get_option('log')])
                self.log_level = BiomajConfig.LOGLEVEL[options.get_option('log')]
                logger.setLevel(self.log_level)
            else:
                hdlr.setLevel(BiomajConfig.LOGLEVEL[self.get('historic.logfile.level')])
                self.log_level = BiomajConfig.LOGLEVEL[self.get('historic.logfile.level')]
                logger.setLevel(self.log_level)
            formatter = logging.Formatter('%(asctime)s %(levelname)-5.5s [%(name)s][%(threadName)s] %(message)s')
            hdlr.setFormatter(formatter)
            logger.addHandler(hdlr)
            '''
            # For monolithic, update default logger/handlers?
            existing_logger_names = list(logging.getLogger().manager.loggerDict.keys())
            for existing_logger in existing_logger_names:
                if existing_logger.startswith('biomaj'):
                    tmp_logger = logging.getLogger(existing_logger)
                    tmp_logger.setLevel(self.log_level)
                    for handler in tmp_logger.handlers:
                        handler.setLevel(self.log_level)
            '''
        else:
            self.log_file = 'none'

        cache_dir = self.get('cache.dir')
        if cache_dir is None:
            print("Configuration file error: cache.dir empty")
            logging.error("cache.dir is not defined")
            sys.exit(1)

        if not os.path.exists(cache_dir):
            os.makedirs(cache_dir)

        process_dir = self.get('process.dir')
        if process_dir is None:
            print("Configuration file error: process.dir empty")
            logging.error("process.dir is not defined")
            sys.exit(1)

        if not os.path.exists(process_dir):
            os.makedirs(process_dir)

        data_dir = self.get('data.dir')
        if data_dir is None:
            print("Configuration file error: data.dir empty")
            logging.error("data.dir is not defined")
            sys.exit(1)

        if not os.path.exists(data_dir):
            os.makedirs(data_dir)

        lock_dir = self.get('lock.dir')
        if lock_dir is None:
            print("Configuration file error: lock.dir empty")
            logging.error("lock.dir is not defined")
            sys.exit(1)

        if not os.path.exists(lock_dir):
            os.makedirs(lock_dir)

    def reset_logger(self):
        '''
        Reconfigure logging
        '''
        if self.log_file == 'none':
            return
        if os.path.exists(os.path.expanduser('~/.biomaj.cfg')):
            logging.config.fileConfig(os.path.expanduser('~/.biomaj.cfg'))
        else:
            logging.config.fileConfig(BiomajConfig.config_file)

        logger = logging.getLogger()
        hdlr = logging.FileHandler(self.log_file)
        hdlr.setLevel(self.log_level)
        formatter = logging.Formatter('%(asctime)s %(levelname)-5.5s [%(name)s][%(threadName)s] %(message)s')
        hdlr.setFormatter(formatter)
        logger.addHandler(hdlr)

    def set(self, prop, value, section='GENERAL'):
        self.config_bank.set(section, prop, value)

    def _in_env(self, prop):
        '''
        Search for env variable BIOMAJ_X_Y_Z for property X.Y.Z
        '''
        env_prop = 'BIOMAJ_' + prop.upper().replace('.', '_')
        if env_prop in os.environ:
            return os.environ[env_prop]
        else:
            return None

    def get_bool(self, prop, section='GENERAL', escape=True, default=None):
        """
        Get a boolean property from bank or general configration. Optionally in section.
        """
        value = None

        if self._in_env(prop):
            value = self._in_env(prop)
        else:
            value = self.get(prop, section, escape, default)

        if value is None:
            return False
        if value is True or value == 'true' or value == '1':
            return True
        else:
            return False

    def get(self, prop, section='GENERAL', escape=True, default=None):
        """
        Get a property from bank or general configration. Optionally in section.
        """
        if self._in_env(prop):
            return self._in_env(prop)
        # Compatibility fields
        if prop == 'depends':
            depend = self.get('db.source', section, escape, None)
            if depend:
                return depend

        if self.config_bank.has_option(section, prop):
            val = self.config_bank.get(section, prop)
            if prop == 'remote.dir' and not val.endswith('/'):
                val = val + '/'
            # If regexp, escape backslashes
            if escape and (prop == 'local.files' or prop == 'remote.files' or
                                   prop == 'http.parse.dir.line' or
                                   prop == 'http.parse.file.line'):
                val = val.replace('\\\\', '\\')
            return val

        if self.user_config is not None:
            if self.user_config.has_option(section, prop):
                return self.user_config.get(section, prop)

        if BiomajConfig.global_config.has_option(section, prop):
            return BiomajConfig.global_config.get(section, prop)

        if prop in BiomajConfig.DEFAULTS:
            return BiomajConfig.DEFAULTS[prop]

        return default

    def get_time(self):
        """
        Return last modification time of config files
        """
        return self.last_modified

    def check(self):
        """
        Check configuration
        """
        self.set('localrelease', '')
        self.set('remoterelease', '')

        status = True
        if not self.get('data.dir'):
            logging.error('data.dir is not set')
            status = False
        if not self.get('conf.dir'):
            logging.error('conf.dir is not set')
            status = False
        if not self.get('log.dir'):
            logging.error('log.dir is not set')
            status = False
        if not self.get('process.dir'):
            logging.error('process.dir is not set')
            status = False
        if not self.get('lock.dir'):
            logging.error('lock.dir is not set')
            status = False
        if not self.get('cache.dir'):
            logging.error('cache.dir is not set')
            status = False
        if not self.get('db.fullname'):
            logging.warn('db.fullname is not set')
        if not self.get('db.formats'):
            logging.warn('db.formats is not set')
        if self.get('use_ldap'):
            if not self.get('ldap.host') or not self.get('ldap.port') or \
                    not self.get('ldap.dn'):
                logging.error('use_ldap set to 1 but missing configuration')
                status = False
        if self.get('use_elastic'):
            if not self.get('elastic_nodes') or not self.get('elastic_index'):
                logging.error('use_elastic set to 1 but missing configuration')
                status = False

        # if not self.get('celery.queue') or not self.get('celery.broker'):
        #    logging.warn('celery config is not set, that\'s fine if you do not use Celery for background tasks')

        if not self.get('mail.smtp.host'):
            logging.warn('SMTP mail config not set, you will not be able to send emails')
        if self.get('mail.smtp.host') and not self.get('mail.from'):
            logging.error('Mail origin mail.from not set')
            status = False

        if not self.get('offline.dir.name'):
            logging.error('offline.dir.name is not set')
            status = False
        elif self.get('offline.dir.name').startswith('/'):
            logging.error('offline dir must be relative to data.dir and should not start with a /')
            status = False
        if not self.get('dir.version'):
            logging.error('dir.version is not set')
            status = False
        elif self.get('dir.version').startswith('/'):
            logging.error('dir.version must be relative to data.dir and should not start with a /')
            status = False
        if not self.get('protocol'):
            logging.error('protocol is not set')
            status = False
        else:
            protocol = self.get('protocol')
            allowed_protocols = ['none', 'multi', 'local', 'ftp', 'sftp', 'http',
                                 'https', 'directftp', 'directhttp', 'directhttps', 'rsync', 'irods']
            if protocol not in allowed_protocols:
                logging.error('Protocol not supported: ' + protocol)
                status = False
            if protocol not in ['multi', 'none']:
                if protocol != 'local' and not self.get('server'):
                    logging.error('server not set')
                    status = False
                if not self.get('remote.dir'):
                    logging.error('remote.dir not set')
                    status = False
                elif not self.get('remote.dir').endswith('/'):
                    logging.error('remote.dir must end with a /')
                    return False
                if protocol not in ['direcftp', 'directhttp', 'directhttps'] and\
                        not self.get('remote.files') and\
                        not self.get('remote.list'):
                    logging.error('remote.files not set')
                    status = False
            if 'irods' in protocol:
                if not self.get('irods.user'):
                    logging.error('irods user is not set')
                    status = False
                if not self.get('irods.password'):
                    logging.error('irods password is not set')
                    status = False
                if not self.get('irods.port'):
                    logging.error('irods port is not set')
                    status = False
                if not self.get('irods.protocol'):
                    logging.error('irods protocol is not set')
                    status = False
        if not self.get('local.files'):
            logging.error('local.files is not set')
            status = False
        # Remove processes
        processes = ['db.remove.process', 'db.pre.process']
        for process in processes:
            if self.get(process):
                metas = self.get(process).split(',')
                for meta in metas:
                    if not self.get(meta):
                        logging.error('Metaprocess ' + meta + ' not defined')
                        status = False
                    else:
                        procs = self.get(meta).split(',')
                        for proc in procs:
                            if not self.get(proc + '.name'):
                                logging.error('Process ' + proc + ' not defined')
                                status = False
                            else:
                                if not self.get(proc + '.exe'):
                                    logging.error('Process exe for ' + proc +
                                                  ' not defined')
                                    status = False
                                if self.get(proc + '.args') is None:
                                    logging.error(
                                        'Process args for ' + proc +
                                        ' not defined')
                                    status = False
        # Check blocks
        if self.get('BLOCKS'):
            blocks = self.get('BLOCKS').split(',')
            for block in blocks:
                if not self.get(block + '.db.post.process'):
                    logging.error('Block ' + block + ' not defined')
                    status = False
                else:
                    metas = self.get(block + '.db.post.process').split(',')
                    for meta in metas:
                        if not self.get(meta):
                            logging.error('Metaprocess ' + meta + ' not defined')
                            status = False
                        else:
                            procs = self.get(meta).split(',')
                            for proc in procs:
                                if not self.get(proc + '.name'):
                                    logging.error('Process ' + proc +
                                                  ' not defined')
                                    status = False
                                else:
                                    if not self.get(proc + '.exe'):
                                        logging.error('Process exe for ' +
                                                      proc + ' not defined')
                                        status = False
                                    if self.get(proc + '.args') is None:
                                        logging.error('Process args for ' +
                                                      proc + ' not defined')
                                        status = False
        return status



import os
import re
import logging
import shutil
import datetime
import time
import subprocess
from subprocess import CalledProcessError
import socket
import pkg_resources
from mimetypes import MimeTypes
import requests


class Utils(object):
    """
    Utility classes
    """

    services = ['USER', 'DOWNLOAD', 'PROCESS', 'RELEASE', 'CRON', 'DAEMON', 'FTP', 'WATCHER']

    mime = None

    @staticmethod
    def get_module_version(module_name):
        '''
        Get module_name version and latest version on pypi
        '''
        try:
            latest = None
            req = requests.get('https://pypi.python.org/pypi/' + module_name + '/json')
            if req.status_code == 200:
                req_json = req.json()
                latest = str(req_json['info']['version'])
            return (pkg_resources.get_distribution(module_name).version, latest)
        except Exception as e:
            logging.debug(str(e))
            return (None, latest)

    @staticmethod
    def get_service_endpoint(config, service):
        '''
        Get endpoint from config for a service. If not defined, return global endpoint
        '''
        if service.upper() not in Utils.services:
            if 'web' not in config or 'local_endpoint' not in config['web']:
                return None
            return config['web']['local_endpoint']
        service_endpoint_name = 'local_endpoint_' + service.lower()
        if service_endpoint_name in config['web'] and config['web'][service_endpoint_name]:
            return config['web'][service_endpoint_name]
        else:
            if 'web' not in config or 'local_endpoint' not in config['web']:
                return None
            return config['web']['local_endpoint']

    @staticmethod
    def service_config_override(config):
        if 'rabbitmq' not in config:
            config['rabbitmq'] = {}
        if 'RABBITMQ_HOST' in os.environ and os.environ['RABBITMQ_HOST']:
            config['rabbitmq']['host'] = os.environ['RABBITMQ_HOST']
        if 'RABBITMQ_PORT' in os.environ and os.environ['RABBITMQ_PORT']:
            config['rabbitmq']['port'] = int(os.environ['RABBITMQ_PORT'])
        if 'RABBITMQ_USER' in os.environ and os.environ['RABBITMQ_USER']:
            config['rabbitmq']['user'] = os.environ['RABBITMQ_USER']
        if 'RABBITMQ_PASSWORD' in os.environ and os.environ['RABBITMQ_PASSWORD']:
            config['rabbitmq']['password'] = os.environ['RABBITMQ_PASSWORD']
        if 'RABBITMQ_VHOST' in os.environ and os.environ['RABBITMQ_VHOST']:
            config['rabbitmq']['virtual_host'] = os.environ['RABBITMQ_VHOST']
        if 'consul' not in config:
            config['consul'] = {'host': None, 'id': None}
        if not config['consul']['id']:
            if 'HOSTNAME' in os.environ and os.environ['HOSTNAME']:
                config['consul']['id'] = os.environ['HOSTNAME']
            else:
                config['consul']['id'] = socket.gethostname()

        if 'web' not in config:
            config['web'] = {}
        if 'BIOMAJ_HOSTNAME' in os.environ and os.environ['BIOMAJ_HOSTNAME']:
            config['web']['hostname'] = os.environ['BIOMAJ_HOSTNAME']
        else:
            if 'HOSTNAME' in os.environ and os.environ['HOSTNAME']:
                config['web']['hostname'] = os.environ['HOSTNAME']
            else:
                config['web']['hostname'] = socket.gethostname()

        if 'docker' not in config:
            config['docker'] = {}
        if 'DOCKER_URL' in os.environ and os.environ['DOCKER_URL']:
            config['docker']['url'] = os.environ['DOCKER_URL']
        else:
            config['docker']['url'] = None

        if 'REDIS_HOST' in os.environ and os.environ['REDIS_HOST']:
            config['redis']['host'] = os.environ['REDIS_HOST']
        if 'REDIS_PORT' in os.environ and os.environ['REDIS_PORT']:
            config['redis']['port'] = int(os.environ['REDIS_PORT'])
        if 'REDIS_PREFIX' in os.environ and os.environ['REDIS_PREFIX']:
            config['redis']['prefix'] = os.environ['REDIS_PREFIX']
        if 'CONSUL_HOST' in os.environ and os.environ['CONSUL_HOST']:
            config['consul']['host'] = os.environ['CONSUL_HOST']
        if 'CONSUL_ID' in os.environ and os.environ['CONSUL_ID']:
            config['consul']['id'] = os.environ['CONSUL_ID']
        if 'WEB_PORT' in os.environ and os.environ['WEB_PORT']:
            config['web']['port'] = int(os.environ['WEB_PORT'])
        if 'WEB_LOCAL_ENDPOINT' in os.environ and os.environ['WEB_LOCAL_ENDPOINT']:
            config['web']['local_endpoint'] = os.environ['WEB_LOCAL_ENDPOINT']
        for service in Utils.services:
            if 'WEB_LOCAL_ENDPOINT_' + service in os.environ and os.environ['WEB_LOCAL_ENDPOINT_' + service]:
                config['web']['local_endpoint_' + service.lower()] = os.environ['WEB_LOCAL_ENDPOINT_' + service]
        if 'mongo' not in config:
            config['mongo'] = {'url': None, 'db': 'biomaj'}
        if 'MONGO_URL' in os.environ and os.environ['MONGO_URL']:
            config['mongo']['url'] = os.environ['MONGO_URL']
        if 'MONGO_DB' in os.environ and os.environ['MONGO_DB']:
            config['mongo']['db'] = os.environ['MONGO_DB']

    @staticmethod
    def get_folder_size(folder):
        """
        Get directory path full size

        :param folder: directory path
        :type folder: str
        """
        if not os.path.exists(folder):
            return -1
        folder_size = 0
        for (path, dirs, files) in os.walk(folder):
            for ffile in files:
                filename = os.path.join(path, ffile)
                folder_size += os.path.getsize(filename)
        return folder_size

    @staticmethod
    def detect_format(filename):
        """
        try to detect file format by extension
        """
        if Utils.mime is None:
            Utils.mime = MimeTypes()
            mimesfile = os.path.join(os.path.dirname(__file__), 'mimes-bio.txt')
            Utils.mime.read(mimesfile, True)
        return Utils.mime.guess_type(filename, True)

    @staticmethod
    def get_more_recent_file(files):
        """
        Return the date of the most recent file in list.

        Each file is a dict like with (at least) parameters: year, month, day
        """
        release = None
        for rfile in files:
            if release is None:
                release = {'year': rfile['year'], 'month': rfile['month'], 'day': rfile['day']}
            else:
                rel_date = datetime.date(int(release['year']), int(release['month']), int(release['day']))
                file_date = datetime.date(int(rfile['year']), int(rfile['month']), int(rfile['day']))
                if file_date > rel_date:
                    release['year'] = rfile['year']
                    release['month'] = rfile['month']
                    release['day'] = rfile['day']
        return release

    @staticmethod
    def month_to_num(date):
        return {
            'Jan': 1,
            'Feb': 2,
            'Mar': 3,
            'Apr': 4,
            'May': 5,
            'Jun': 6,
            'Jul': 7,
            'Aug': 8,
            'Sep': 9,
            'Oct': 10,
            'Nov': 11,
            'Dec': 12,
            '01': 1,
            '02': 2,
            '03': 3,
            '04': 4,
            '05': 5,
            '06': 6,
            '07': 7,
            '08': 8,
            '09': 9,
            '10': 10,
            '11': 11,
            '12': 12
            }[date]

    @staticmethod
    def copy_files(files_to_copy, to_dir, move=False, lock=None):
        """
        Copy or move files to to_dir, keeping directory structure.

        Copy keeps the original file stats.
        Files should have attributes name and root:
        - root: root directory
        - name: relative path of file in root directory

        /root/file/file1 will be copied in to_dir/file/file1

        :param files_to_copy: list of files to copy
        :type files_to_copy: list
        :param to_dir: destination directory
        :type to_dir: str
        :param move: move instead of copy
        :type move: bool
        :param lock: thread lock object for multi-threads
        :type lock: Lock
        """
        logger = logging.getLogger('biomaj')
        nb_files = len(files_to_copy)
        cur_files = 1
        for file_to_copy in files_to_copy:
            logger.debug(str(cur_files) + '/' + str(nb_files) + ' copy file ' + file_to_copy['name'])
            cur_files += 1
            from_file = file_to_copy['root'] + '/' + file_to_copy['name']
            to_file = to_dir + '/' + file_to_copy['name']
            if lock is not None:
                lock.acquire()
                try:
                    if not os.path.exists(os.path.dirname(to_file)):
                        os.makedirs(os.path.dirname(to_file))
                except Exception as e:
                    logger.error(e)
                finally:
                    lock.release()

            else:
                if not os.path.exists(os.path.dirname(to_file)):
                    try:
                        os.makedirs(os.path.dirname(to_file))
                    except Exception as e:
                        logger.error(e)
            if move:
                shutil.move(from_file, to_file)
            else:
                start_time = datetime.datetime.now()
                start_time = time.mktime(start_time.timetuple())
                shutil.copyfile(from_file, to_file)
                end_time = datetime.datetime.now()
                end_time = time.mktime(end_time.timetuple())
                file_to_copy['download_time'] = end_time - start_time
                shutil.copystat(from_file, to_file)

    @staticmethod
    def copy_files_with_regexp(from_dir, to_dir, regexps, move=False, lock=None):
        """
        Copy or move files from from_dir to to_dir matching regexps.
        Copy keeps the original file stats.

        :param from_dir: origin directory
        :type from_dir: str
        :param to_dir: destination directory
        :type to_dir: str
        :param regexps: list of regular expressions that files in from_dir should match to be copied
        :type regexps: list
        :param move: move instead of copy
        :type move: bool
        :param lock: thread lock object for multi-threads
        :type lock: Lock
        :return: list of copied files with their size
        """
        logger = logging.getLogger('biomaj')
        files_to_copy = []
        for root, dirs, files in os.walk(from_dir, topdown=True):
            for name in files:
                for reg in regexps:
                    file_relative_path = os.path.join(root, name).replace(from_dir, '')
                    if file_relative_path.startswith('/'):
                        file_relative_path = file_relative_path.replace('/', '', 1)
                    if reg == "**/*":
                        files_to_copy.append({'name': file_relative_path})
                        continue
                    if re.match(reg, file_relative_path):
                        files_to_copy.append({'name': file_relative_path})
                        continue

        for file_to_copy in files_to_copy:
            from_file = from_dir + '/' + file_to_copy['name']
            to_file = to_dir + '/' + file_to_copy['name']

            if lock is not None:
                lock.acquire()
                try:
                    if not os.path.exists(os.path.dirname(to_file)):
                        os.makedirs(os.path.dirname(to_file))
                except Exception as e:
                    logger.error(e)
                finally:
                    lock.release()
            else:
                if not os.path.exists(os.path.dirname(to_file)):
                    os.makedirs(os.path.dirname(to_file))
            if not os.path.exists(from_file):
                logger.warn('File does not exists: %s' % (from_file))
                continue
            if move:
                shutil.move(from_file, to_file)
            else:
                shutil.copyfile(from_file, to_file)
                shutil.copystat(from_file, to_file)
            file_to_copy['size'] = os.path.getsize(to_file)
            f_stat = datetime.datetime.fromtimestamp(os.path.getmtime(to_file))
            file_to_copy['year'] = str(f_stat.year)
            file_to_copy['month'] = str(f_stat.month)
            file_to_copy['day'] = str(f_stat.day)
            (file_format, encoding) = Utils.detect_format(to_file)
            file_to_copy['format'] = file_format
        return files_to_copy

    @staticmethod
    def archive_check(archivefile):
        """
        Test file archive integrity

        :param file: full path to file to check and uncompress
        :type file: str
        :return: True if ok, False if an error occured
        """
        logger = logging.getLogger('biomaj')
        try:
            if archivefile.endswith('.tar.gz'):
                subprocess.check_call("tar tfz " + archivefile, shell=True)
            elif archivefile.endswith('.tar'):
                subprocess.check_call("tar tf " + archivefile, shell=True)
            elif archivefile.endswith('.bz2'):
                subprocess.check_call("tar tjf " + archivefile, shell=True)
            elif archivefile.endswith('.gz'):
                subprocess.check_call("gunzip -t " + archivefile, shell=True)
            elif archivefile.endswith('.zip'):
                subprocess.check_call("unzip -t " + archivefile, shell=True)
        except CalledProcessError as uncompresserror:
            logger.error("Archive integrity error of %s: %s" % (archivefile, str(uncompresserror)))
            return False

        return True

    @staticmethod
    def uncompress(archivefile, remove=True):
        """
        Test if file is an archive, and uncompress it
        Remove archive file if specified

        :param file: full path to file to check and uncompress
        :type file: str
        :param remove: remove archive if present
        :type remove: bool
        :return: True if ok, False if an error occured
        """
        is_archive = False
        logger = logging.getLogger('biomaj')
        try:
            if archivefile.endswith('.tar.gz'):
                subprocess.check_call("tar xfz " + archivefile + " --overwrite -C " + os.path.dirname(archivefile), shell=True)
                is_archive = True
            elif archivefile.endswith('.tar'):
                subprocess.check_call("tar xf " + archivefile + " --overwrite -C " + os.path.dirname(archivefile), shell=True)
                is_archive = True
            elif archivefile.endswith('.bz2'):
                subprocess.check_call("tar xjf " + archivefile + " --overwrite -C " + os.path.dirname(archivefile), shell=True)
                is_archive = True
            elif archivefile.endswith('.gz'):
                subprocess.check_call("gunzip -f " + archivefile, shell=True)
                is_archive = True
            elif archivefile.endswith('.zip'):
                subprocess.check_call("unzip -o " + archivefile + " -d " + os.path.dirname(archivefile), shell=True)
                is_archive = True
        except CalledProcessError as uncompresserror:
            logger.error("Uncompress error of %s: %s" % (archivefile, str(uncompresserror)))
            return False

        if is_archive:
            logger.debug('Uncompress:uncompress:' + archivefile)

        if is_archive and remove and os.path.exists(archivefile):
            os.remove(archivefile)

        return True
